/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.formcycle;

import static de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.*;

import java.io.ByteArrayOutputStream;
import java.util.Collections;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.http.converter.protobuf.ProtobufHttpMessageConverter;
import org.springframework.mock.web.MockMultipartFile;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;
import org.springframework.test.web.servlet.setup.MockMvcBuilders;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.FormHeader;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroupTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.formdata.ServiceKonto.PostfachAddress;
import de.ozgcloud.eingang.common.formdata.ServiceKontoTestFactory;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.common.formdata.StringBasedIdentifier;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import lombok.SneakyThrows;

class FormDataControllerTest {

	@Spy
	@InjectMocks
	private FormDataController controller;

	@Mock
	private FormCycleFormDataMapper mapper;
	@Mock
	private SemantikAdapter semantikAdapter;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;
	@Mock
	private FormDataHtmlCleaner htmlCleaner;

	private MockMvc mockMvc;

	@BeforeEach
	void init() {
		mockMvc = MockMvcBuilders.standaloneSetup(controller)
				.setMessageConverters(new ProtobufHttpMessageConverter())
				.build();
	}

	@Nested
	class ReceiveFormData {

		static final String VORGANG_NUMMER = "VorgangNummer";

		private FormData mappedFormData = FormDataTestFactory.create();

		@Captor
		private ArgumentCaptor<FormData> formDataCaptor;

		@BeforeEach
		void init() {
			when(htmlCleaner.clean(any())).thenReturn(mappedFormData);
			when(mapper.toFormData(any())).thenReturn(mappedFormData);
			when(vorgangNummerSupplier.get()).thenReturn(VORGANG_NUMMER);
		}

		@Test
		void shouldReturnSuccess() throws Exception {
			doPostRequest().andExpect(status().is2xxSuccessful());
		}

		@Test
		void shouldCallVorgangNummerSupplier() {
			doPostRequest();

			verify(vorgangNummerSupplier).get();
		}

		@Test
		@SneakyThrows
		void shouldResponseWithVorgangNummer() {

			var confirmation = FormCycleConfirmationResponse.parseFrom(
					doPostRequest().andReturn().getResponse().getContentAsByteArray());

			assertThat(confirmation.getVorgangNummer()).isEqualTo(VORGANG_NUMMER);
		}

		@Test
		void shouldSetVorgangNummer() {
			doPostRequest();

			verify(semantikAdapter).processFormData(formDataCaptor.capture());
			assertThat(formDataCaptor.getValue().getHeader().getRequestId()).isEqualTo(VORGANG_NUMMER);
		}

		@Test
		void shouldCallHtmlCleaner() {
			doPostRequest();

			verify(htmlCleaner).clean(any());
		}

		@Test
		void shouldCallMapper() {
			doPostRequest();

			verify(mapper).toFormData(notNull());
		}

		@Test
		void shouldCallSemantikAdapter() {
			doPostRequest();

			verify(semantikAdapter).processFormData(formDataCaptor.capture());
			assertThat(formDataCaptor.getValue()).usingRecursiveComparison()
					.ignoringFields("representations", "attachments", "numberOfAttachments", "header.requestId")
					.isEqualTo(mappedFormData);
		}

		@Test
		void shouldMapPostkorbId() {
			doPostRequest();

			verify(semantikAdapter).processFormData(formDataCaptor.capture());
			assertThat(getPostfachIdFormData(formDataCaptor.getValue())).isEqualTo(FormCyclePostfachAddressTestFactory.POSTKORB_ID);
		}

		@Nested
		class Representations {

			@BeforeEach
			void init() {
				var formData = FormDataTestFactory.createBuilder().clearRepresentations().numberOfRepresentations(0).build();
				when(mapper.toFormData(any())).thenReturn(formData);
				when(htmlCleaner.clean(any())).thenReturn(formData);
			}

			@Test
			void shouldBePresent() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				var formData = formDataCaptor.getValue();
				assertThat(formData.getRepresentations()).hasSize(1);
				assertThat(formData.getNumberOfRepresentations()).isEqualTo(1);
			}

			@Test
			void shouldBeFilled() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				assertThat(formDataCaptor.getValue().getRepresentations()).first().usingRecursiveComparison()
						.ignoringFields("id", "vendorId", "file")
						.isEqualTo(IncomingFileTestFactory.create());
			}

			@Test
			@SneakyThrows
			void shouldBeFineWithoutRepresentations() {
				mockMvc.perform(
						multipart("/formData")
								.file(new MockMultipartFile("formData", null, FormDataController.HTTP_TYPE_PROTOBUF, buildTestFormData())))
						.andExpect(status().isOk());
			}
		}

		@Nested
		class Attachments {
			@BeforeEach
			void init() {
				var formData = FormDataTestFactory.createBuilder().clearAttachments().numberOfAttachments(0).build();
				when(mapper.toFormData(any())).thenReturn(formData);
				when(htmlCleaner.clean(any())).thenReturn(formData);
			}

			@Test
			void shouldHaveGroup() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				assertThat(formDataCaptor.getValue().getAttachments()).hasSize(1);
			}

			@Test
			void shouldSetNumberOfAttachments() {
				doPostRequest();

				verify(semantikAdapter).processFormData(formDataCaptor.capture());
				assertThat(formDataCaptor.getValue().getNumberOfAttachments()).isEqualTo(1);
			}

		}

		@SneakyThrows
		private ResultActions doPostRequest() {
			return mockMvc.perform(
					multipart("/formData")
							.file(new MockMultipartFile("formData", null, FormDataController.HTTP_TYPE_PROTOBUF, buildTestFormData()))
							.file(IncomingFileTestFactory.asMultipartFile("representations"))
							.file(asMultipartFile("attachments",
									createBuilder().name(IncomingFileGroupTestFactory.VENDOR_ID_XXX).build())));
		}
	}

	@SneakyThrows
	static byte[] buildTestFormData() {
		return buildTestFormData(FormCycleFormDataTestFactory.create());
	}

	@SneakyThrows
	static byte[] buildTestFormData(FormCycleFormData formData) {
		var out = new ByteArrayOutputStream();
		formData.writeTo(out);
		return out.toByteArray();
	}

	@Nested
	class TestAttachmentGroupsBuilder {

		private FormDataController.AttachmentGroupsBuilder groupsBuilder;

		@BeforeEach
		void init() {
			groupsBuilder = controller.new AttachmentGroupsBuilder(
					Collections.singleton(FormCycleAttachmentGroupTestFactory.create()),
					Optional.of(Collections.singleton(IncomingFileTestFactory.asMultipartFile(NAME))));
		}

		@Nested
		class TestBuildGroup {
			@Test
			void shouldSetName() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.create());

				assertThat(group.getName()).isEqualTo(IncomingFileGroupTestFactory.NAME);
			}

			@Test
			void shouldHaveFile() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.create());

				assertThat(group.getFiles()).hasSize(1).first()
						.usingRecursiveComparison().ignoringFields("id", "vendorId", "file")
						.isEqualTo(IncomingFileTestFactory.create());
			}

			@Test
			void shouldBeFineWithoutFile() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.createBuilder().clearFileId().build());

				assertThat(group.getFiles()).isEmpty();
			}

			@Test
			void shouldBeFineWithMissingFile() {
				var group = groupsBuilder.buildGroup(FormCycleAttachmentGroupTestFactory.createBuilder().clearFileId().addFileId("missing").build());

				assertThat(group.getFiles()).isEmpty();
			}
		}
	}

	@Nested
	class TestServiceKontoMapping {

		@Test
		void shouldMapServiceKontoType() {
			var formData = controller.addServiceKonto(FormCycleFormDataTestFactory.create(), buildEmptyFormDataWithHeader());

			assertThat(getServiceKontoType(formData)).isEqualTo(FormCycleServiceKontoTestFactory.TYPE);
		}

		@Test
		void shouldNotMapEmptyServiceKonto() {
			controller.addServiceKonto(FormCycleFormData.newBuilder().build(), buildEmptyFormDataWithHeader());

			verify(controller, never()).buildServiceKonto(any());
		}

		@Test
		void shouldMapTrustLevel() {
			var formData = controller.addServiceKonto(FormCycleFormDataTestFactory.create(), buildEmptyFormDataWithHeader());

			assertThat(formData.getHeader().getServiceKonto().getTrustLevel()).isEqualTo(ServiceKontoTestFactory.TRUST_LEVEL);
		}

		String getServiceKontoType(FormData formData) {
			return formData.getHeader().getServiceKonto().getType();
		}

		@Nested
		class TestPostkorbIdMapping {

			@Test
			void shouldMapPostkorbId() {
				var formData = controller.addServiceKonto(FormCycleFormDataTestFactory.create(), buildEmptyFormDataWithHeader());

				assertThat(getPostfachIdFormData(formData)).isEqualTo(FormCyclePostfachAddressTestFactory.POSTKORB_ID);
			}

			@Test
			void shouldNotMapEmptyPostkorbId() {
				var postfachAddress = controller.buildPostfachAddress(FormCycleServiceKonto.newBuilder().build());

				assertThat(postfachAddress).isNull();
			}

			@Test
			void shouldMapPostfachIdentifier() {
				var postfachAddress = controller.buildPostfachAddress(FormCycleServiceKontoTestFactory.create());

				assertThat(getIdentifier(postfachAddress)).isEqualTo(FormCyclePostfachAddressTestFactory.POSTKORB_ID);
			}

			private String getIdentifier(PostfachAddress postfachAddress) {
				return ((StringBasedIdentifier) postfachAddress.getIdentifier()).getPostfachId();
			}

			@Test
			void shouldMapPostfachAddressVersion() {
				var postfachAddress = controller.buildPostfachAddress(FormCycleServiceKontoTestFactory.create());

				assertThat(postfachAddress.getVersion()).isEqualTo(FormCyclePostfachAddressTestFactory.VERSION);
			}
		}

		private FormData buildEmptyFormDataWithHeader() {
			return FormData.builder().header(FormHeader.builder().build()).build();
		}
	}

	private String getPostfachIdFormData(FormData formData) {
		return ((StringBasedIdentifier) formData.getHeader().getServiceKonto().getPostfachAddresses()
				.get(0).getIdentifier()).getPostfachId();
	}
}
